#include <stdio.h>
#include <stdlib.h>

#define MAX_BUFF_SIZE 30+1

typedef struct list_st {
	double value;
	struct list_st *next;
} LIST;

void init_list(LIST **head) {
	*head = NULL;
}

void add_to_list(LIST *new, LIST **head) {
	if(*head == NULL) { // List is empty
		*head = new;
	} else { // It's actually easier to add/remove items from start than from the end of the list ;)
		new->next = *head;
		*head = new;
	}
}

LIST *create_new_item(double value) {
	LIST *new = (LIST *)malloc(sizeof(LIST));
	if (new == NULL) {
		printf("Not enough RAM!\n");
		exit(21);
	}

	new->value = value;
	new->next = NULL;

	return new;
}

void push(LIST **head, double value) {
	LIST *new = create_new_item(value);
	add_to_list(new, head);
}

double pop(LIST **head) {
	if (*head == NULL) {
		printf("Tried to POP an empty stack!\n");
		exit(22);
	}

	LIST *first = *head;
	double value = first->value;

	*head = first->next;
	free(first);

	return value;
}

FILE *safe_fopen(char *filename, char *mode, int error_code) {
	FILE *fp = fopen(filename, mode);
	if (fp == NULL) {
		printf("Can't open '%s'!\n", filename);
		exit(error_code);
	}
	return fp;
}

double calc(FILE *in, LIST **head) {
	char token[MAX_BUFF_SIZE];

	while(fscanf(in, "%s", token) != EOF) {
		char ch = token[0];
		if(ch >= '0' && ch <= '9' ) { // Maybe it's a number (aka operand)?
			push(head, atof(token));
		} else { // Maybe it's an operator?
			char operator = ch;
			double op2 = pop(head);
			double op1 = pop(head);

			double result;
			switch (operator) {
				case '+': result = op1 + op2; break;
				case '-': result = op1 - op2; break;
				case '*': result = op1 * op2; break;
				case '/': result = op1 / op2; break; // A word of advice: divide by zero check might be useful here
				default:
					printf("Unknown operator '%c'!\n", operator);
					exit(31);
			}

			push(head, result);
		}
	}

	return pop(head); // The end result should be on top of the stack
}

int main(int arg_num, char *args[]) {
	if (arg_num != 2) {
		printf("USAGE: %s IN_FILENAME\n", args[0]);
		exit(11);
	}

	char *in_filename = args[1];
	FILE *in = safe_fopen(in_filename, "r", 1);

	LIST *head;
	init_list(&head);

	printf("Result: %f\n", calc(in, &head));

	if (head != NULL) { // Sanity check
		printf("Stack should be empty after all calculations have been completed, but it's not!\n");
		exit(32);
	}

	fclose(in);

	return 0;
}
